#pragma once

/// iir_filter_biquad
/// 
/// Biquad form IIR (from fm10Gen.c)
/// @brief Perform IIR filtering sample by sample on doubles.
/// Implements Biquad form calculations.
///	@param[in] input New input sample
///	@param[in] *coef Pointer to filter coefficient data with size 4*n + 1 (gain)
///	@param[in] n Number of second order sections in filter definition
///	@param[in,out] *history Pointer to filter history data of size 2*n
///	@return Result of filter calculation
inline double
iir_filter_biquad(double input, const double* coef, int n, double* history);

/// biquad
/// 
/// Calculates one set of cascaded second-order sections to form an IIR filter
/// using normal floating point operations. Filter runs once.
/// 
/// @param inp Input value
/// @param out Output value
/// @param coeff Filter coefficienst arranged in LIGO format
/// @param hist History values
/// @param sections Number of second-order sections in the filter
inline
void biquad(const double* inp, double* out, const double* coeff, double* hist, int sections);

/// biquadAVX2
/// 
/// Calculates 4 sets of cascaded second-order sections to form an IIR filter
/// using AVX2 operations. Filter runs once.
/// 
/// All double arrays passed in need to be 32 byte aligned.
/// 
/// @param inp Input values arranged 4 wide
/// @param out Output values arranged 4 wide
/// @param coeff Filter coefficienst arranged in LIGO format and 4 wide each
/// @param hist History values arranged 8 wide
/// @param sections Number of second-order sections in the filter
inline
void biquadAVX2(const double* inp, double* out, const double* coeff, double* hist, int sections);

/// biquadAVX2double
/// 
/// Calculates 8 sets of cascaded second-order sections to form an IIR filter
/// using AVX2 operations. Filter runs twice internally.
/// This function can run on the same data structures as biquadAVX512.
/// 
/// All double arrays passed in need to be 32 byte aligned.
/// 
/// @param inp Input values arranged 8 wide
/// @param out Output values arranged 8 wide
/// @param coeff Filter coefficienst arranged in LIGO format and 4 wide each
/// @param hist History values arranged 16 wide
/// @param sections Number of second-order sections in the filter
inline
void biquadAVX2double(const double* inp, double* out, const double* coeff, double* hist, int sections);

#ifdef AVX_512_SUPPORT
/// biquadAVX512
/// 
/// Calculates 8 sets of cascaded second-order sections to form an IIR filter
/// using AVX-512 operations. Filter runs once.
/// 
/// All double arrays passed in need to be 64 byte aligned.
/// 
/// @param inp Input values arranged 8 wide
/// @param out Output values arranged 8 wide
/// @param coeff Filter coefficienst arranged in LIGO format and 4 wide each
/// @param hist History values arranged 16 wide
/// @param sections Number of second-order sections in the filter
inline
void biquadAVX512(const double* inp, double* out, const double* coeff, double* hist, int sections);
#endif

/// biquad_stride
/// 
/// Calculates a set of cascaded second-order sections to form an IIR filter.
/// 
/// @param inp Input values arranged stride wide and samples deep
/// @param out Output values arranged stride wide
/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
/// @param hist History values arranged double stride wide
/// @param sections Number of second-order sections in the filter
/// @param stride Width of data array
/// @param samples Number of samples in input array which are decimated down
inline
void biquad_stride1(const double* inp, double* out, const double* coeff, double* hist, int sections, int stride = 1, int samples = 1);

inline
void biquad_stride2(const double* inp, double* out, const double* coeff, double* hist, int sections, int stride = 2, int samples = 1);

inline
void biquad_stride4(const double* inp, double* out, const double* coeff, double* hist, int sections, int stride = 4, int samples = 1);

inline
void biquad_avx2_stride4_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride = 4, int samples = 1);

inline
void biquad_avx2_stride8_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride = 8, int samples = 1);

inline
void biquad_avx2_stride16_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride = 16, int samples = 1);

inline
void biquad_avx2_stride32_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride = 32, int samples = 1);

#ifdef AVX_512_SUPPORT
inline
void biquad_avx512_stride8_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride = 8, int samples = 1);

inline
void biquad_avx512_stride16_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride = 16, int samples = 1);

inline
void biquad_avx512_stride32_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride = 32, int samples = 1);
#endif

#include "BiquadInline.h"