#pragma once
#ifdef __GNUC__
#include <x86intrin.h>
#else
#include <intrin.h>
#endif


inline double
iir_filter_biquad(double input, const double* coef, int n, double* history)
{

	int     i;
	const double* coef_ptr;
	double* hist1_ptr, * hist2_ptr;
	double  output, new_w, new_u, w, u, a11, a12, c1, c2;

	coef_ptr = coef; /* coefficient pointer */

	hist1_ptr = history; /* first history */
	hist2_ptr = hist1_ptr + 1; /* next history */

	output = input * (*coef_ptr++); /* overall input scale factor */

	for (i = 0; i < n; i++)
	{

		w = *hist1_ptr;
		u = *hist2_ptr;

		a11 = *coef_ptr++;
		a12 = *coef_ptr++;
		c1 = *coef_ptr++;
		c2 = *coef_ptr++;

		new_w = output + a11 * w + a12 * u;
		output = output + w * c1 + u * c2;
		new_u = w + u;

		*hist1_ptr++ = new_w;
		*hist2_ptr++ = new_u;
		hist1_ptr++;
		hist2_ptr++;
	}

	return (output);
}


inline
void biquad(const double* inp, double* out, const double* coeff, double* hist, int sections)
{
	int j;		// loops over filter taps
	double y;	// filter output value
	double u, w;// current history values
#ifndef LIGO_BIQUAD
	double x;	// filter input value
#endif

	// get first input value and mulitply it with filter gain
	y = coeff[0] * inp[0];
	// Loop over sections
	for (j = 0; j < sections; ++j)
	{
		// Fetch previous history
		w = hist[2 * j + 0];
		u = hist[2 * j + 1];

#ifdef LIGO_BIQUAD
		// Calculate and save new history
		hist[2 * j + 0] = y + coeff[4 * j + 1] * w + coeff[4 * j + 2] * u;
		// Difference equations for LIGO biquad
		y = y + coeff[4 * j + 3] * w + coeff[4 * j + 4] * u;
		// Calculate and save new history
		hist[2 * j + 1] = u + w;
#else
		// Difference equations for transposed direct form II biquad
		x = y;
		y = x + w;
		// Calculate and save new history
		hist[2 * j + 0] = u + coeff[4 * j + 3] * x - coeff[4 * j + 1] * y;
		hist[2 * j + 1] = coeff[4 * j + 4] * x - coeff[4 * j + 2] * y;
#endif
	}
	// Store output
	out[0] = y;
}

inline
void biquadAVX2(const double* inp, double* out, const double* coeff, double* hist, int sections)
{
	int j;	// loops over filter taps

	__m256d y;	// filter value
	__m256d u, w;// current history values
#ifndef LIGO_BIQUAD
	__m256d x;	// filter input value
#endif
	// typecast double pointers into __m256d pointers
	const __m256d* const minp = (__m256d*)inp;
	__m256d* const mout = (__m256d*)out;
	const __m256d* const mcoeff = (__m256d*)coeff;
	__m256d* const mhist = (__m256d*)hist;

	// get first input value and mulitply it with filter gain

	y = _mm256_mul_pd(mcoeff[0], *minp);

	// Loop over sections
	for (j = 0; j < sections; ++j)
	{
		// Fetch coefficients
		// Fetch previous history
		w = mhist[2 * j + 0];
		u = mhist[2 * j + 1];

#ifdef LIGO_BIQUAD
		// Calculate and save new history
		mhist[2 * j + 0] = _mm256_fmadd_pd(mcoeff[4 * j + 2], u, _mm256_fmadd_pd(mcoeff[4 * j + 1], w, y));
		// Difference equations for LIGO biquad
		y = _mm256_fmadd_pd(mcoeff[4 * j + 4], u, _mm256_fmadd_pd(mcoeff[4 * j + 3], w, y));
		// Calculate and save new history
		mhist[2 * j + 1] = _mm256_add_pd(u, w);
#else
		// Difference equations for transposed direct form II biquad
		x = y;
		y = _mm256_add_pd(x, w);
		// Calculate and save new history
		u = _mm256_fmadd_pd(mcoeff[4 * j + 3], x, u);
		mhist[2 * j + 0] = _mm256_fnmadd_pd(mcoeff[4 * j + 1], y, u);
		mhist[2 * j + 1] = _mm256_fnmadd_pd(mcoeff[4 * j + 2], y, _mm256_mul_pd(mcoeff[4 * j + 4], x));
#endif
	}

	// Store output
	*mout = y;
}

inline
void biquadAVX2double(const double* inp, double* out, const double* coeff, double* hist, int sections)
{
	int j;	// loops over filter taps

	__m256d y;	// filter value
	__m256d u;
	__m256d w; // current history values
#ifndef LIGO_BIQUAD
	__m256d x;	// filter input value
#endif
	// typecast double pointers into __m256d pointers
	const __m256d* const minp = (__m256d*)inp;
	__m256d* const mout = (__m256d*)out;
	const __m256d* const mcoeff = (__m256d*)coeff;
	__m256d* const mhist = (__m256d*)hist;

	// get first input value and mulitply it with filter gain

	y = _mm256_mul_pd(mcoeff[0], minp[0]);

	// Loop over sections
	for (j = 0; j < sections; ++j)
	{
		// Fetch coefficients
		// Fetch previous history
		w = mhist[4 * j + 0];
		u = mhist[4 * j + 2];

#ifdef LIGO_BIQUAD
		// Calculate and save new history
		mhist[4 * j + 0] = _mm256_fmadd_pd(mcoeff[8 * j + 4], u, _mm256_fmadd_pd(mcoeff[8 * j + 2], w, y));
		// Difference equations for LIGO biquad
		y = _mm256_fmadd_pd(mcoeff[8 * j + 8], u, _mm256_fmadd_pd(mcoeff[8 * j + 6], w, y));
		// Calculate and save new history
		mhist[4 * j + 2] = _mm256_add_pd(u, w);
#else
		// Difference equations for transposed direct form II biquad
		x = y;
		y = _mm256_add_pd(x, w);
		// Calculate and save new history
		u = _mm256_fmadd_pd(mcoeff[8 * j + 6], x, u);
		mhist[4 * j + 0] = _mm256_fnmadd_pd(mcoeff[8 * j + 2], y, u);
		mhist[4 * j + 2] = _mm256_fnmadd_pd(mcoeff[8 * j + 4], y, _mm256_mul_pd(mcoeff[8 * j + 8], x));
#endif
	}

	// Store output
	mout[0] = y;

	y = _mm256_mul_pd(mcoeff[1], minp[1]);

	// Loop over sections
	for (j = 0; j < sections; ++j)
	{
		// Fetch coefficients
		// Fetch previous history
		w = mhist[4 * j + 1];
		u = mhist[4 * j + 3];

#ifdef LIGO_BIQUAD
		// Calculate and save new history
		mhist[4 * j + 1] = _mm256_fmadd_pd(mcoeff[8 * j + 5], u, _mm256_fmadd_pd(mcoeff[8 * j + 3], w, y));
		// Difference equations for LIGO biquad
		y = _mm256_fmadd_pd(mcoeff[8 * j + 9], u, _mm256_fmadd_pd(mcoeff[8 * j + 7], w, y));
		// Calculate and save new history
		mhist[4 * j + 3] = _mm256_add_pd(u, w);
#else
		// Difference equations for transposed direct form II biquad
		x = y;
		y = _mm256_add_pd(x, w);
		// Calculate and save new history
		u = _mm256_fmadd_pd(mcoeff[8 * j + 7], x, u);
		mhist[4 * j + 1] = _mm256_fnmadd_pd(mcoeff[8 * j + 3], y, u);
		mhist[4 * j + 3] = _mm256_fnmadd_pd(mcoeff[8 * j + 5], y, _mm256_mul_pd(mcoeff[8 * j + 9], x));
#endif
	}

	// Store output
	mout[1] = y;
}

#ifdef AVX_512_SUPPORT
inline
void biquadAVX512(const double* inp, double* out, const double* coeff, double* hist, int sections)
{
	int j;	// loops over filter taps

	__m512d y;	// filter value
	__m512d u, w;// current history values
#ifndef LIGO_BIQUAD
	__m512d x;	// filter input value
#endif
	// typecast double pointers into __m512d pointers
	const __m512d* const minp = (__m512d*)inp;
	__m512d* const mout = (__m512d*)out;
	const __m512d* const mcoeff = (__m512d*)coeff;
	__m512d* const mhist = (__m512d*)hist;

	// get first input value and mulitply it with filter gain

	y = _mm512_mul_pd(mcoeff[0], *minp);

	// Loop over sections
	for (j = 0; j < sections; ++j)
	{
		// Fetch coefficients
		// Fetch previous history
		w = mhist[2 * j + 0];
		u = mhist[2 * j + 1];

#ifdef LIGO_BIQUAD
		// Calculate and save new history
		mhist[2 * j + 0] = _mm512_fmadd_pd(mcoeff[4 * j + 2], u, _mm512_fmadd_pd(mcoeff[4 * j + 1], w, y));
		// Difference equations for LIGO biquad
		y = _mm512_fmadd_pd(mcoeff[4 * j + 4], u, _mm512_fmadd_pd(mcoeff[4 * j + 3], w, y));
		// Calculate and save new history
		mhist[2 * j + 1] = _mm512_add_pd(u, w);
#else
		// Difference equations for transposed direct form II biquad
		x = y;
		y = _mm512_add_pd(x, w);
		// Calculate and save new history
		u = _mm512_fmadd_pd(mcoeff[4 * j + 3], x, u);
		mhist[2 * j + 0] = _mm512_fnmadd_pd(mcoeff[4 * j + 1], y, u);
		mhist[2 * j + 1] = _mm512_fnmadd_pd(mcoeff[4 * j + 2], y, _mm512_mul_pd(mcoeff[4 * j + 4], x));
#endif
	}

	// Store output
	*mout = y;
}
#endif

inline
void biquad_stride1(const double* inp, double* out, const double* coeff, double* hist, int sections, int stride, int samples)
{
	int i;		// loops over stride width
	int n;		// loops over number of samples
	int j;		// loops over filter taps
	double y;	// filter output value
	double u, w;// current history values
#ifndef LIGO_BIQUAD
	double x;	// filter input value
#endif

	if (samples < 1) return;
	if (stride < 1) return;

	// loop over stride
	for (i = 0; i < stride; ++i)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y = coeff[0] * inp[i + n * stride];
			// Loop over sections
			for (j = 0; j < sections; ++j)
			{
				// Fetch previous history
				w = hist[i + (2 * j + 0) * stride];
				u = hist[i + (2 * j + 1) * stride];

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				hist[i + (2 * j + 0) * stride] = y + coeff[4 * j + 1] * w + coeff[4 * j + 2] * u;
				// Difference equations for LIGO biquad
				y = y + coeff[4 * j + 3] * w + coeff[4 * j + 4] * u;
				// Calculate and save new history
				hist[i + (2 * j + 1) * stride] = u + w;
#else
				// Difference equations for transposed direct form II biquad
				x = y;
				y = x + w;
				// Calculate and save new history
				hist[i + (2 * j + 0) * stride] = u + coeff[4 * j + 3] * x - coeff[4 * j + 1] * y;
				hist[i + (2 * j + 1) * stride] = coeff[4 * j + 4] * x - coeff[4 * j + 2] * y;
#endif
			}
		}
		// Store output
		out[i] = y;
	}
}

inline
void biquad_stride2(const double* inp, double* out, const double* coeff, double* hist, int sections, int stride, int samples)
{
	// intrinsic stride width is 2
#define stride2_intrinsic 2

	int i;		// loops over stride width
	int n;		// loops over number of samples
	int j;		// loops over filter taps
	double y1, y2;	// filter output value
	double u1, u2;
	double w1, w2;// current history values
#ifndef LIGO_BIQUAD
	double x1, x2;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride2_intrinsic) || (stride % stride2_intrinsic != 0)) return;

	// loop over stride
	for (i = 0; i + stride2_intrinsic <= stride; i = i + stride2_intrinsic)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y1 = coeff[0] * inp[i + n * stride];
			y2 = coeff[0] * inp[i + n * stride + 1];
			// Loop over sections
			for (j = 0; j < sections; ++j)
			{
				// Fetch previous history
				w1 = hist[i + (2 * j + 0) * stride];
				w2 = hist[i + (2 * j + 0) * stride + 1];
				u1 = hist[i + (2 * j + 1) * stride];
				u2 = hist[i + (2 * j + 1) * stride + 1];

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				hist[i + (2 * j + 0) * stride + 0] = y1 + coeff[4 * j + 1] * w1 + coeff[4 * j + 2] * u1;
				hist[i + (2 * j + 0) * stride + 1] = y2 + coeff[4 * j + 1] * w2 + coeff[4 * j + 2] * u2;
				// Difference equations for LIGO biquad
				y1 = y1 + coeff[4 * j + 3] * w1 + coeff[4 * j + 4] * u1;
				y2 = y2 + coeff[4 * j + 3] * w2 + coeff[4 * j + 4] * u2;
				// Calculate and save new history
				hist[i + (2 * j + 1) * stride + 0] = u1 + w1;
				hist[i + (2 * j + 1) * stride + 1] = u2 + w2;
#else
				// Difference equations for transposed direct form II biquad
				x1 = y1;
				x2 = y2;
				y1 = x1 + w1;
				y2 = x2 + w2;
				// Calculate and save new history
				hist[i + (2 * j + 0) * stride + 0] = u1 + coeff[4 * j + 3] * x1 - coeff[4 * j + 1] * y1;
				hist[i + (2 * j + 0) * stride + 1] = u2 + coeff[4 * j + 3] * x2 - coeff[4 * j + 1] * y2;
				hist[i + (2 * j + 1) * stride + 0] = coeff[4 * j + 4] * x1 - coeff[4 * j + 2] * y1;
				hist[i + (2 * j + 1) * stride + 1] = coeff[4 * j + 4] * x2 - coeff[4 * j + 2] * y2;
#endif
			}
		}
		// Store output
		out[i + 0] = y1;
		out[i + 1] = y2;
	}
}

inline
void biquad_stride4(const double* inp, double* out, const double* coeff, double* hist, int sections, int stride, int samples)
{
	// intrinsic stride width is 4
#define stride4_intrinsic 4

	int i;		// loops over stride width
	int n;		// loops over number of samples
	int j;		// loops over filter taps
	double y1, y2, y3, y4;	// filter output value
	double u1, u2, u3, u4;
	double w1, w2, w3, w4;// current history values
#ifndef LIGO_BIQUAD
	double x1, x2, x3, x4;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride4_intrinsic) || (stride % stride4_intrinsic != 0)) return;

	// loop over stride
	for (i = 0; i + stride4_intrinsic <= stride; i = i + stride4_intrinsic)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y1 = coeff[0] * inp[i + n * stride];
			y2 = coeff[0] * inp[i + n * stride + 1];
			y3 = coeff[0] * inp[i + n * stride + 2];
			y4 = coeff[0] * inp[i + n * stride + 3];
			// Loop over sections
			for (j = 0; j < sections; ++j)
			{
				// Fetch previous history
				w1 = hist[i + (2 * j + 0) * stride];
				w2 = hist[i + (2 * j + 0) * stride + 1];
				w3 = hist[i + (2 * j + 0) * stride + 2];
				w4 = hist[i + (2 * j + 0) * stride + 3];
				u1 = hist[i + (2 * j + 1) * stride];
				u2 = hist[i + (2 * j + 1) * stride + 1];
				u3 = hist[i + (2 * j + 1) * stride + 2];
				u4 = hist[i + (2 * j + 1) * stride + 3];

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				hist[i + (2 * j + 0) * stride + 0] = y1 + coeff[4 * j + 1] * w1 + coeff[4 * j + 2] * u1;
				hist[i + (2 * j + 0) * stride + 1] = y2 + coeff[4 * j + 1] * w2 + coeff[4 * j + 2] * u2;
				hist[i + (2 * j + 0) * stride + 2] = y3 + coeff[4 * j + 1] * w3 + coeff[4 * j + 2] * u3;
				hist[i + (2 * j + 0) * stride + 3] = y4 + coeff[4 * j + 1] * w4 + coeff[4 * j + 2] * u4;
				// Difference equations for LIGO biquad
				y1 = y1 + coeff[4 * j + 3] * w1 + coeff[4 * j + 4] * u1;
				y2 = y2 + coeff[4 * j + 3] * w2 + coeff[4 * j + 4] * u2;
				y3 = y3 + coeff[4 * j + 3] * w3 + coeff[4 * j + 4] * u3;
				y4 = y4 + coeff[4 * j + 3] * w4 + coeff[4 * j + 4] * u4;
				// Calculate and save new history
				hist[i + (2 * j + 1) * stride + 0] = u1 + w1;
				hist[i + (2 * j + 1) * stride + 1] = u2 + w2;
				hist[i + (2 * j + 1) * stride + 2] = u3 + w3;
				hist[i + (2 * j + 1) * stride + 3] = u4 + w4;
#else
				// Difference equations for transposed direct form II biquad
				x1 = y1;
				x2 = y2;
				x3 = y3;
				x4 = y4;
				y1 = x1 + w1;
				y2 = x2 + w2;
				y3 = x3 + w3;
				y4 = x4 + w4;
				// Calculate and save new history
				hist[i + (2 * j + 0) * stride + 0] = u1 + coeff[4 * j + 3] * x1 - coeff[4 * j + 1] * y1;
				hist[i + (2 * j + 0) * stride + 1] = u2 + coeff[4 * j + 3] * x2 - coeff[4 * j + 1] * y2;
				hist[i + (2 * j + 0) * stride + 2] = u3 + coeff[4 * j + 3] * x3 - coeff[4 * j + 1] * y3;
				hist[i + (2 * j + 0) * stride + 3] = u4 + coeff[4 * j + 3] * x4 - coeff[4 * j + 1] * y4;

				hist[i + (2 * j + 1) * stride + 0] = coeff[4 * j + 4] * x1 - coeff[4 * j + 2] * y1;
				hist[i + (2 * j + 1) * stride + 1] = coeff[4 * j + 4] * x2 - coeff[4 * j + 2] * y2;
				hist[i + (2 * j + 1) * stride + 2] = coeff[4 * j + 4] * x3 - coeff[4 * j + 2] * y3;
				hist[i + (2 * j + 1) * stride + 3] = coeff[4 * j + 4] * x4 - coeff[4 * j + 2] * y4;
#endif
			}
		}
		// Store output
		out[i + 0] = y1;
		out[i + 1] = y2;
		out[i + 2] = y3;
		out[i + 3] = y4;
	}
}

inline
void biquad_avx2_stride4_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride, int samples)
{
	// intrinsic stride width is 4
#define stride_mm256 4
#define stride_mm256_q(i) (	 (i == 0) ? 0 * stride_mm256: \
									 (i == 1) ? 1 * stride_mm256 : \
									 (i == 2) ? 2 * stride_mm256 : \
									 (i == 3) ? 3 * stride_mm256 : \
									 (i == 4) ? 4 * stride_mm256 : \
									 (i == 5) ? 5 * stride_mm256 : \
									 (i == 6) ? 6 * stride_mm256 : \
									 7 * stride_mm256)
#define stride4_mm256 4
#define stride4_mm256_num (stride4_mm256 / stride_mm256)

	// number of sections are 3
#define sections3_intrinsic 3

	size_t i;	// loops over stride width
	int n;		// loops over number of samples
	size_t j;	// loops over filter taps
	int k;      // loops over coefficients

	__m256d y;	// filter value
	__m256d u, w;// current history values
#ifndef LIGO_BIQUAD
	__m256d x;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride4_mm256) || (stride % stride4_mm256 != 0)) return;

	// load coefficients into AVX registers
	__m256d fcoeff[sections3_intrinsic * 4 + 1];
	for (k = 0; k < sections3_intrinsic * 4 + 1; ++k)
	{
		fcoeff[k] = _mm256_set1_pd(coeff[k]);
	}
	// loop over stride
	for (i = 0; i <= stride - stride4_mm256; i = i + stride4_mm256)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride));
			// Loop over sections
			for (j = 0; j < sections3_intrinsic; ++j)
			{
				// Fetch previous history
				w = *(__m256d*)(hist + i + (2 * j + 0) * stride);
				u = *(__m256d*)(hist + i + (2 * j + 1) * stride);

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				* (__m256d*)(hist + i + (2 * j + 0) * stride) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u, _mm256_fmadd_pd(fcoeff[4 * j + 1], w, y));
				// Difference equations for LIGO biquad
				y = _mm256_fmadd_pd(fcoeff[4 * j + 4], u, _mm256_fmadd_pd(fcoeff[4 * j + 3], w, y));
				// Calculate and save new history
				*(__m256d*)(hist + i + (2 * j + 1) * stride) = _mm256_add_pd(u, w);
#else
				// Difference equations for transposed direct form II biquad
				x = y;
				y = _mm256_add_pd(x, w);
				// Calculate and save new history
				u = _mm256_fmadd_pd(fcoeff[4 * j + 3], x, u);
				*(__m256d*)(hist + i + (2 * j + 0) * stride) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y, u);
				*(__m256d*)(hist + i + (2 * j + 1) * stride) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y, _mm256_mul_pd(fcoeff[4 * j + 4], x));
#endif
			}
		}
		// Store output
		*((__m256d*)(out + i)) = y;
	}

}

inline
void biquad_avx2_stride8_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride, int samples)
{
	// intrinsic stride width is 8
#define stride_mm256 4
#define stride_mm256_q(i) (	 (i == 0) ? 0 * stride_mm256: \
									 (i == 1) ? 1 * stride_mm256 : \
									 (i == 2) ? 2 * stride_mm256 : \
									 (i == 3) ? 3 * stride_mm256 : \
									 (i == 4) ? 4 * stride_mm256 : \
									 (i == 5) ? 5 * stride_mm256 : \
									 (i == 6) ? 6 * stride_mm256 : \
									 7 * stride_mm256)
#define stride8_mm256 8
#define stride8_mm256_num (stride8_mm256 / stride_mm256)

	// number of sections are 3
#define sections3_intrinsic 3

	size_t i;	// loops over stride width
	int n;		// loops over number of samples
	size_t j;	// loops over filter taps
	int k;      // loops over coefficients

	__m256d y1, y2;	// filter value
	__m256d u1, u2;
	__m256d w1, w2;// current history values
#ifndef LIGO_BIQUAD
	__m256d x1, x2;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride8_mm256) || (stride % stride8_mm256 != 0)) return;

	// load coefficients into AVX registers
	__m256d fcoeff[sections3_intrinsic * 4 + 1];
	for (k = 0; k < sections3_intrinsic * 4 + 1; ++k)
	{
		fcoeff[k] = _mm256_set1_pd(coeff[k]);;
	}
	// loop over stride
	for (i = 0; i <= stride - stride8_mm256; i = i + stride8_mm256)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y1 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(0)));
			y2 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(1)));
			// Loop over sections
			for (j = 0; j < sections3_intrinsic; ++j)
			{
				// Fetch previous history
				w1 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0));
				w2 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1));
				u1 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0));
				u2 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1));

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u1, _mm256_fmadd_pd(fcoeff[4 * j + 1], w1, y1));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u2, _mm256_fmadd_pd(fcoeff[4 * j + 1], w2, y2));
				// Difference equations for LIGO biquad
				y1 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u1, _mm256_fmadd_pd(fcoeff[4 * j + 3], w1, y1));
				y2 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u2, _mm256_fmadd_pd(fcoeff[4 * j + 3], w2, y2));
				// Calculate and save new history
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0)) = _mm256_add_pd(u1, w1);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1)) = _mm256_add_pd(u2, w2);
#else
				// Difference equations for transposed direct form II biquad
				x1 = y1;
				x2 = y2;
				y1 = _mm256_add_pd(x1, w1);
				y2 = _mm256_add_pd(x2, w2);
				// Calculate and save new history
				u1 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x1, u1);
				u2 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x2, u2);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y1, u1);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y2, u2);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y1, _mm256_mul_pd(fcoeff[4 * j + 4], x1));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y2, _mm256_mul_pd(fcoeff[4 * j + 4], x2));
#endif
			}
		}
		// Store output
		*((__m256d*)(out + i + stride_mm256_q(0))) = y1;
		*((__m256d*)(out + i + stride_mm256_q(1))) = y2;
	}

}

inline
void biquad_avx2_stride16_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride, int samples)
{
	// intrinsic stride width is 16
#define stride_mm256 4
#define stride_mm256_q(i) (	 (i == 0) ? 0 * stride_mm256: \
									 (i == 1) ? 1 * stride_mm256 : \
									 (i == 2) ? 2 * stride_mm256 : \
									 (i == 3) ? 3 * stride_mm256 : \
									 (i == 4) ? 4 * stride_mm256 : \
									 (i == 5) ? 5 * stride_mm256 : \
									 (i == 6) ? 6 * stride_mm256 : \
									 7 * stride_mm256)
#define stride16_mm256 16
#define stride16_mm256_num (stride16_mm256 / stride_mm256)

	// number of sections are 3
#define sections3_intrinsic 3

	size_t i;	// loops over stride width
	int n;		// loops over number of samples
	size_t j;	// loops over filter taps
	int k;      // loops over coefficients

	__m256d y1, y2, y3, y4;	// filter value
	__m256d u1, u2, u3, u4;
	__m256d w1, w2, w3, w4;// current history values
#ifndef LIGO_BIQUAD
	__m256d x1, x2, x3, x4;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride16_mm256) || (stride % stride16_mm256 != 0)) return;

	// load coefficients into AVX registers
	__m256d fcoeff[sections3_intrinsic * 4 + 1];
	for (k = 0; k < sections3_intrinsic * 4 + 1; ++k)
	{
		fcoeff[k] = _mm256_set1_pd(coeff[k]);
	}
	// loop over stride
	for (i = 0; i <= stride - stride16_mm256; i = i + stride16_mm256)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y1 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(0)));
			y2 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(1)));
			y3 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(2)));
			y4 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(3)));
			// Loop over sections
			for (j = 0; j < sections3_intrinsic; ++j)
			{
				// Fetch previous history
				w1 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0));
				w2 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1));
				w3 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(2));
				w4 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(3));
				u1 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0));
				u2 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1));
				u3 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(2));
				u4 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(3));

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u1, _mm256_fmadd_pd(fcoeff[4 * j + 1], w1, y1));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u2, _mm256_fmadd_pd(fcoeff[4 * j + 1], w2, y2));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(2)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u3, _mm256_fmadd_pd(fcoeff[4 * j + 1], w3, y3));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(3)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u4, _mm256_fmadd_pd(fcoeff[4 * j + 1], w4, y4));
				// Difference equations for LIGO biquad
				y1 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u1, _mm256_fmadd_pd(fcoeff[4 * j + 3], w1, y1));
				y2 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u2, _mm256_fmadd_pd(fcoeff[4 * j + 3], w2, y2));
				y3 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u3, _mm256_fmadd_pd(fcoeff[4 * j + 3], w3, y3));
				y4 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u4, _mm256_fmadd_pd(fcoeff[4 * j + 3], w4, y4));
				// Calculate and save new history
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0)) = _mm256_add_pd(u1, w1);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1)) = _mm256_add_pd(u2, w2);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(2)) = _mm256_add_pd(u3, w3);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(3)) = _mm256_add_pd(u4, w4);
#else
				// Difference equations for transposed direct form II biquad
				x1 = y1;
				x2 = y2;
				x3 = y3;
				x4 = y4;
				y1 = _mm256_add_pd(x1, w1);
				y2 = _mm256_add_pd(x2, w2);
				y3 = _mm256_add_pd(x3, w3);
				y4 = _mm256_add_pd(x4, w4);
				// Calculate and save new history
				u1 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x1, u1);
				u2 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x2, u2);
				u3 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x3, u3);
				u4 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x4, u4);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y1, u1);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y2, u2);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(2)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y3, u3);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(3)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y4, u4);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y1, _mm256_mul_pd(fcoeff[4 * j + 4], x1));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y2, _mm256_mul_pd(fcoeff[4 * j + 4], x2));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(2)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y3, _mm256_mul_pd(fcoeff[4 * j + 4], x3));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(3)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y4, _mm256_mul_pd(fcoeff[4 * j + 4], x4));
#endif
			}
		}
		// Store output
		*((__m256d*)(out + i + stride_mm256_q(0))) = y1;
		*((__m256d*)(out + i + stride_mm256_q(1))) = y2;
		*((__m256d*)(out + i + stride_mm256_q(2))) = y3;
		*((__m256d*)(out + i + stride_mm256_q(3))) = y4;
	}
}

inline
void biquad_avx2_stride32_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride, int samples)
{
	// intrinsic stride width is 32
#define stride_mm256 4
#define stride_mm256_q(i) (	 (i == 0) ? 0 * stride_mm256: \
									 (i == 1) ? 1 * stride_mm256 : \
									 (i == 2) ? 2 * stride_mm256 : \
									 (i == 3) ? 3 * stride_mm256 : \
									 (i == 4) ? 4 * stride_mm256 : \
									 (i == 5) ? 5 * stride_mm256 : \
									 (i == 6) ? 6 * stride_mm256 : \
									 7 * stride_mm256)
#define stride32_mm256 32
#define stride32_mm256_num (stride32_mm256 / stride_mm256)

	// number of sections are 3
#define sections3_intrinsic 3

	size_t i;	// loops over stride width
	int n;		// loops over number of samples
	size_t j;	// loops over filter taps
	int k;      // loops over coefficients

	__m256d y1, y2, y3, y4, y5, y6, y7, y8;	// filter value
	__m256d u1, u2, u3, u4, u5, u6, u7, u8;
	__m256d w1, w2, w3, w4, w5, w6, w7, w8; // current history values
#ifndef LIGO_BIQUAD
	__m256d x1, x2, x3, x4, x5, x6, x7, x8;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride32_mm256) || (stride % stride32_mm256 != 0)) return;

	// load coefficients into AVX registers
	__m256d fcoeff[sections3_intrinsic * 4 + 1];
	for (k = 0; k < sections3_intrinsic * 4 + 1; ++k)
	{
		fcoeff[k] = _mm256_set1_pd(coeff[k]);
	}
	// loop over stride
	for (i = 0; i <= stride - stride32_mm256; i = i + stride32_mm256)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y1 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(0)));
			y2 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(1)));
			y3 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(2)));
			y4 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(3)));
			y5 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(4)));
			y6 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(5)));
			y7 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(6)));
			y8 = _mm256_mul_pd(fcoeff[0], *(__m256d*)(inp + i + n * stride + stride_mm256_q(7)));
			// Loop over sections
			for (j = 0; j < sections3_intrinsic; ++j)
			{
				// Fetch previous history
				w1 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0));
				w2 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1));
				w3 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(2));
				w4 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(3));
				w5 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(4));
				w6 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(5));
				w7 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(6));
				w8 = *(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(7));
				u1 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0));
				u2 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1));
				u3 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(2));
				u4 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(3));
				u5 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(4));
				u6 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(5));
				u7 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(6));
				u8 = *(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(7));

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u1, _mm256_fmadd_pd(fcoeff[4 * j + 1], w1, y1));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u2, _mm256_fmadd_pd(fcoeff[4 * j + 1], w2, y2));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(2)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u3, _mm256_fmadd_pd(fcoeff[4 * j + 1], w3, y3));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(3)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u4, _mm256_fmadd_pd(fcoeff[4 * j + 1], w4, y4));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(4)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u5, _mm256_fmadd_pd(fcoeff[4 * j + 1], w5, y5));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(5)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u6, _mm256_fmadd_pd(fcoeff[4 * j + 1], w6, y6));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(6)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u7, _mm256_fmadd_pd(fcoeff[4 * j + 1], w7, y7));
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(7)) = _mm256_fmadd_pd(fcoeff[4 * j + 2], u8, _mm256_fmadd_pd(fcoeff[4 * j + 1], w8, y8));
				// Difference equations for LIGO biquad
				y1 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u1, _mm256_fmadd_pd(fcoeff[4 * j + 3], w1, y1));
				y2 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u2, _mm256_fmadd_pd(fcoeff[4 * j + 3], w2, y2));
				y3 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u3, _mm256_fmadd_pd(fcoeff[4 * j + 3], w3, y3));
				y4 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u4, _mm256_fmadd_pd(fcoeff[4 * j + 3], w4, y4));
				y5 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u5, _mm256_fmadd_pd(fcoeff[4 * j + 3], w5, y5));
				y6 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u6, _mm256_fmadd_pd(fcoeff[4 * j + 3], w6, y6));
				y7 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u7, _mm256_fmadd_pd(fcoeff[4 * j + 3], w7, y7));
				y8 = _mm256_fmadd_pd(fcoeff[4 * j + 4], u8, _mm256_fmadd_pd(fcoeff[4 * j + 3], w8, y8));
				// Calculate and save new history
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0)) = _mm256_add_pd(u1, w1);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1)) = _mm256_add_pd(u2, w2);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(2)) = _mm256_add_pd(u3, w3);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(3)) = _mm256_add_pd(u4, w4);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(4)) = _mm256_add_pd(u5, w5);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(5)) = _mm256_add_pd(u6, w6);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(6)) = _mm256_add_pd(u7, w7);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(7)) = _mm256_add_pd(u8, w8);
#else
				// Difference equations for transposed direct form II biquad
				x1 = y1;
				x2 = y2;
				x3 = y3;
				x4 = y4;
				x5 = y5;
				x6 = y6;
				x7 = y7;
				x8 = y8;
				y1 = _mm256_add_pd(x1, w1);
				y2 = _mm256_add_pd(x2, w2);
				y3 = _mm256_add_pd(x3, w3);
				y4 = _mm256_add_pd(x4, w4);
				y5 = _mm256_add_pd(x5, w5);
				y6 = _mm256_add_pd(x6, w6);
				y7 = _mm256_add_pd(x7, w7);
				y8 = _mm256_add_pd(x8, w8);
				// Calculate and save new history
				u1 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x1, u1);
				u2 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x2, u2);
				u3 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x3, u3);
				u4 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x4, u4);
				u5 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x5, u5);
				u6 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x6, u6);
				u7 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x7, u7);
				u8 = _mm256_fmadd_pd(fcoeff[4 * j + 3], x8, u8);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(0)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y1, u1);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(1)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y2, u2);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(2)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y3, u3);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(3)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y4, u4);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(4)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y5, u5);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(5)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y6, u6);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(6)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y7, u7);
				*(__m256d*)(hist + i + (2 * j + 0) * stride + stride_mm256_q(7)) = _mm256_fnmadd_pd(fcoeff[4 * j + 1], y8, u8);
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(0)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y1, _mm256_mul_pd(fcoeff[4 * j + 4], x1));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(1)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y2, _mm256_mul_pd(fcoeff[4 * j + 4], x2));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(2)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y3, _mm256_mul_pd(fcoeff[4 * j + 4], x3));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(3)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y4, _mm256_mul_pd(fcoeff[4 * j + 4], x4));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(4)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y5, _mm256_mul_pd(fcoeff[4 * j + 4], x5));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(5)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y6, _mm256_mul_pd(fcoeff[4 * j + 4], x6));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(6)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y7, _mm256_mul_pd(fcoeff[4 * j + 4], x7));
				*(__m256d*)(hist + i + (2 * j + 1) * stride + stride_mm256_q(7)) = _mm256_fnmadd_pd(fcoeff[4 * j + 2], y8, _mm256_mul_pd(fcoeff[4 * j + 4], x8));
#endif
			}
		}
		// Store output
		*((__m256d*)(out + i + stride_mm256_q(0))) = y1;
		*((__m256d*)(out + i + stride_mm256_q(1))) = y2;
		*((__m256d*)(out + i + stride_mm256_q(2))) = y3;
		*((__m256d*)(out + i + stride_mm256_q(3))) = y4;
		*((__m256d*)(out + i + stride_mm256_q(4))) = y5;
		*((__m256d*)(out + i + stride_mm256_q(5))) = y6;
		*((__m256d*)(out + i + stride_mm256_q(6))) = y7;
		*((__m256d*)(out + i + stride_mm256_q(7))) = y8;
	}
}

#ifdef AVX_512_SUPPORT
inline
void biquad_avx512_stride8_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride, int samples)
{
	// intrinsic stride width is 8
#define stride_mm512 8
#define stride_mm512_q(i) (	 (i == 0) ? 0 * stride_mm512: \
									 (i == 1) ? 1 * stride_mm512 : \
									 (i == 2) ? 2 * stride_mm512 : \
									 (i == 3) ? 3 * stride_mm512 : \
									 (i == 4) ? 4 * stride_mm512 : \
									 (i == 5) ? 5 * stride_mm512 : \
									 (i == 6) ? 6 * stride_mm512 : \
									 7 * stride_mm512)
#define stride8_mm512 8
#define stride8_mm152_num (stride8_mm512 / stride_mm512)

	// number of sections are 3
#define sections3_intrinsic 3

	size_t i;	// loops over stride width
	int n;		// loops over number of samples
	size_t j;	// loops over filter taps
	int k;      // loops over coefficients

	__m512d y;	// filter value
	__m512d u, w;// current history values
#ifndef LIGO_BIQUAD
	__m512d x;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride8_mm512) || (stride % stride8_mm512 != 0)) return;

	// load coefficients into AVX registers
	__m512d fcoeff[sections3_intrinsic * 4 + 1];
	for (k = 0; k < sections3_intrinsic * 4 + 1; ++k)
	{
		fcoeff[k] = _mm512_set1_pd(coeff[k]);
	}
	// loop over stride
	for (i = 0; i + stride8_mm512 <= stride; i = i + stride8_mm512)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y = _mm512_mul_pd(fcoeff[0], *(__m512d*)(inp + i + n * stride));
			// Loop over sections
			for (j = 0; j < sections3_intrinsic; ++j)
			{
				// Fetch previous history
				w = *(__m512d*)(hist + i + (2 * j + 0) * stride);
				u = *(__m512d*)(hist + i + (2 * j + 1) * stride);

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				* (__m512d*)(hist + i + (2 * j + 0) * stride) = _mm512_fmadd_pd(fcoeff[4 * j + 2], u, _mm512_fmadd_pd(fcoeff[4 * j + 1], w, y));
				// Difference equations for LIGO biquad
				y = _mm512_fmadd_pd(fcoeff[4 * j + 4], u, _mm512_fmadd_pd(fcoeff[4 * j + 3], w, y));
				// Calculate and save new history
				*(__m512d*)(hist + i + (2 * j + 1) * stride) = _mm512_add_pd(u, w);
#else
				// Difference equations for transposed direct form II biquad
				x = y;
				y = _mm512_add_pd(x, w);
				// Calculate and save new history
				u = _mm512_fmadd_pd(fcoeff[4 * j + 3], x, u);
				*(__m512d*)(hist + i + (2 * j + 0) * stride) = _mm512_fnmadd_pd(fcoeff[4 * j + 1], y, u);
				*(__m512d*)(hist + i + (2 * j + 1) * stride) = _mm512_fnmadd_pd(fcoeff[4 * j + 2], y, _mm512_mul_pd(fcoeff[4 * j + 4], x));
#endif
			}
		}
		// Store output
		*((__m512d*)(out + i)) = y;
	}
}

inline
void biquad_avx512_stride16_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride, int samples)
{
	// intrinsic stride width is 16
#define stride_mm512 8
#define stride_mm512_q(i) (	 (i == 0) ? 0 * stride_mm512: \
									 (i == 1) ? 1 * stride_mm512 : \
									 (i == 2) ? 2 * stride_mm512 : \
									 (i == 3) ? 3 * stride_mm512 : \
									 (i == 4) ? 4 * stride_mm512 : \
									 (i == 5) ? 5 * stride_mm512 : \
									 (i == 6) ? 6 * stride_mm512 : \
									 7 * stride_mm512)
#define stride16_mm512 16
#define stride16_mm152_num (stride16_mm512 / stride_mm512)

	// number of sections are 3
#define sections3_intrinsic 3

	size_t i;	// loops over stride width
	int n;		// loops over number of samples
	size_t j;	// loops over filter taps
	int k;      // loops over coefficients

	__m512d y1, y2;	// filter value
	__m512d u1, u2;
	__m512d w1, w2;// current history values
#ifndef LIGO_BIQUAD
	__m512d x1, x2;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride16_mm512) || (stride % stride16_mm512 != 0)) return;

	// load coefficients into AVX registers
	__m512d fcoeff[sections3_intrinsic * 4 + 1];
	for (k = 0; k < sections3_intrinsic * 4 + 1; ++k)
	{
		fcoeff[k] = _mm512_set1_pd(coeff[k]);;
	}
	// loop over stride
	for (i = 0; i + stride16_mm512 <= stride; i = i + stride16_mm512)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y1 = _mm512_mul_pd(fcoeff[0], *(__m512d*)(inp + i + n * stride + stride_mm512_q(0)));
			y2 = _mm512_mul_pd(fcoeff[0], *(__m512d*)(inp + i + n * stride + stride_mm512_q(1)));
			// Loop over sections
			for (j = 0; j < sections3_intrinsic; ++j)
			{
				// Fetch previous history
				w1 = *(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(0));
				w2 = *(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(1));
				u1 = *(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(0));
				u2 = *(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(1));

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(0)) = _mm512_fmadd_pd(fcoeff[4 * j + 2], u1, _mm512_fmadd_pd(fcoeff[4 * j + 1], w1, y1));
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(1)) = _mm512_fmadd_pd(fcoeff[4 * j + 2], u2, _mm512_fmadd_pd(fcoeff[4 * j + 1], w2, y2));
				// Difference equations for LIGO biquad
				y1 = _mm512_fmadd_pd(fcoeff[4 * j + 4], u1, _mm512_fmadd_pd(fcoeff[4 * j + 3], w1, y1));
				y2 = _mm512_fmadd_pd(fcoeff[4 * j + 4], u2, _mm512_fmadd_pd(fcoeff[4 * j + 3], w2, y2));
				// Calculate and save new history
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(0)) = _mm512_add_pd(u1, w1);
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(1)) = _mm512_add_pd(u2, w2);
#else
				// Difference equations for transposed direct form II biquad
				x1 = y1;
				x2 = y2;
				y1 = _mm512_add_pd(x1, w1);
				y2 = _mm512_add_pd(x2, w2);
				// Calculate and save new history
				u1 = _mm512_fmadd_pd(fcoeff[4 * j + 3], x1, u1);
				u2 = _mm512_fmadd_pd(fcoeff[4 * j + 3], x2, u2);
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(0)) = _mm512_fnmadd_pd(fcoeff[4 * j + 1], y1, u1);
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(1)) = _mm512_fnmadd_pd(fcoeff[4 * j + 1], y2, u2);
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(0)) = _mm512_fnmadd_pd(fcoeff[4 * j + 2], y1, _mm512_mul_pd(fcoeff[4 * j + 4], x1));
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(1)) = _mm512_fnmadd_pd(fcoeff[4 * j + 2], y2, _mm512_mul_pd(fcoeff[4 * j + 4], x2));
#endif
			}
		}
		// Store output
		*((__m512d*)(out + i + stride_mm512_q(0))) = y1;
		*((__m512d*)(out + i + stride_mm512_q(1))) = y2;
	}
}

inline
void biquad_avx512_stride32_section3(const double* inp, double* out, const double* coeff, double* hist, size_t stride, int samples)
{
	// intrinsic stride width is 32
#define stride_mm512 8
#define stride_mm512_q(i) (	 (i == 0) ? 0 * stride_mm512: \
									 (i == 1) ? 1 * stride_mm512 : \
									 (i == 2) ? 2 * stride_mm512 : \
									 (i == 3) ? 3 * stride_mm512 : \
									 (i == 4) ? 4 * stride_mm512 : \
									 (i == 5) ? 5 * stride_mm512 : \
									 (i == 6) ? 6 * stride_mm512 : \
									 7 * stride_mm512)
#define stride32_mm512 32
#define stride32_mm512_num (stride32_mm512 / stride_mm512)

	// number of sections are 3
#define sections3_intrinsic 3

	size_t i;	// loops over stride width
	int n;		// loops over number of samples
	size_t j;	// loops over filter taps
	int k;      // loops over coefficients

	__m512d y1, y2, y3, y4;	// filter value
	__m512d u1, u2, u3, u4;
	__m512d w1, w2, w3, w4;// current history values
#ifndef LIGO_BIQUAD
	__m512d x1, x2, x3, x4;	// filter input value
#endif

	if (samples < 1) return;
	if ((stride < stride32_mm512) || (stride % stride32_mm512 != 0)) return;

	// load coefficients into AVX registers
	__m512d fcoeff[sections3_intrinsic * 4 + 1];
	for (k = 0; k < sections3_intrinsic * 4 + 1; ++k)
	{
		fcoeff[k] = _mm512_set1_pd(coeff[k]);
	}
	// loop over stride
	for (i = 0; i + stride32_mm512 <= stride; i = i + stride32_mm512)
	{
		// loop over number of samples
		for (n = 0; n < samples; ++n)
		{
			// get first input value and mulitply it with filter gain
			y1 = _mm512_mul_pd(fcoeff[0], *(__m512d*)(inp + i + n * stride + stride_mm512_q(0)));
			y2 = _mm512_mul_pd(fcoeff[0], *(__m512d*)(inp + i + n * stride + stride_mm512_q(1)));
			y3 = _mm512_mul_pd(fcoeff[0], *(__m512d*)(inp + i + n * stride + stride_mm512_q(2)));
			y4 = _mm512_mul_pd(fcoeff[0], *(__m512d*)(inp + i + n * stride + stride_mm512_q(3)));
			// Loop over sections
			for (j = 0; j < sections3_intrinsic; ++j)
			{
				// Fetch previous history
				w1 = *(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(0));
				w2 = *(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(1));
				w3 = *(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(2));
				w4 = *(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(3));
				u1 = *(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(0));
				u2 = *(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(1));
				u3 = *(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(2));
				u4 = *(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(3));

#ifdef LIGO_BIQUAD
				// Calculate and save new history
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(0)) = _mm512_fmadd_pd(fcoeff[4 * j + 2], u1, _mm512_fmadd_pd(fcoeff[4 * j + 1], w1, y1));
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(1)) = _mm512_fmadd_pd(fcoeff[4 * j + 2], u2, _mm512_fmadd_pd(fcoeff[4 * j + 1], w2, y2));
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(2)) = _mm512_fmadd_pd(fcoeff[4 * j + 2], u3, _mm512_fmadd_pd(fcoeff[4 * j + 1], w3, y3));
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(3)) = _mm512_fmadd_pd(fcoeff[4 * j + 2], u4, _mm512_fmadd_pd(fcoeff[4 * j + 1], w4, y4));
				// Difference equations for LIGO biquad
				y1 = _mm512_fmadd_pd(fcoeff[4 * j + 4], u1, _mm512_fmadd_pd(fcoeff[4 * j + 3], w1, y1));
				y2 = _mm512_fmadd_pd(fcoeff[4 * j + 4], u2, _mm512_fmadd_pd(fcoeff[4 * j + 3], w2, y2));
				y3 = _mm512_fmadd_pd(fcoeff[4 * j + 4], u3, _mm512_fmadd_pd(fcoeff[4 * j + 3], w3, y3));
				y4 = _mm512_fmadd_pd(fcoeff[4 * j + 4], u4, _mm512_fmadd_pd(fcoeff[4 * j + 3], w4, y4));
				// Calculate and save new history
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(0)) = _mm512_add_pd(u1, w1);
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(1)) = _mm512_add_pd(u2, w2);
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(2)) = _mm512_add_pd(u3, w3);
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(3)) = _mm512_add_pd(u4, w4);
#else
				// Difference equations for transposed direct form II biquad
				x1 = y1;
				x2 = y2;
				x3 = y3;
				x4 = y4;
				y1 = _mm512_add_pd(x1, w1);
				y2 = _mm512_add_pd(x2, w2);
				y3 = _mm512_add_pd(x3, w3);
				y4 = _mm512_add_pd(x4, w4);
				// Calculate and save new history
				u1 = _mm512_fmadd_pd(fcoeff[4 * j + 3], x1, u1);
				u2 = _mm512_fmadd_pd(fcoeff[4 * j + 3], x2, u2);
				u3 = _mm512_fmadd_pd(fcoeff[4 * j + 3], x3, u3);
				u4 = _mm512_fmadd_pd(fcoeff[4 * j + 3], x4, u4);
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(0)) = _mm512_fnmadd_pd(fcoeff[4 * j + 1], y1, u1);
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(1)) = _mm512_fnmadd_pd(fcoeff[4 * j + 1], y2, u2);
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(2)) = _mm512_fnmadd_pd(fcoeff[4 * j + 1], y3, u3);
				*(__m512d*)(hist + i + (2 * j + 0) * stride + stride_mm512_q(3)) = _mm512_fnmadd_pd(fcoeff[4 * j + 1], y4, u4);
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(0)) = _mm512_fnmadd_pd(fcoeff[4 * j + 2], y1, _mm512_mul_pd(fcoeff[4 * j + 4], x1));
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(1)) = _mm512_fnmadd_pd(fcoeff[4 * j + 2], y2, _mm512_mul_pd(fcoeff[4 * j + 4], x2));
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(2)) = _mm512_fnmadd_pd(fcoeff[4 * j + 2], y3, _mm512_mul_pd(fcoeff[4 * j + 4], x3));
				*(__m512d*)(hist + i + (2 * j + 1) * stride + stride_mm512_q(3)) = _mm512_fnmadd_pd(fcoeff[4 * j + 2], y4, _mm512_mul_pd(fcoeff[4 * j + 4], x4));
#endif
			}
		}
		// Store output
		*((__m512d*)(out + i + stride_mm512_q(0))) = y1;
		*((__m512d*)(out + i + stride_mm512_q(1))) = y2;
		*((__m512d*)(out + i + stride_mm512_q(2))) = y3;
		*((__m512d*)(out + i + stride_mm512_q(3))) = y4;
	}
}

#endif
