#pragma once
#include <array>

#ifdef __GNUC__
#define alignit(n) __attribute__((aligned(n)))
#else
#define alignit(n) __declspec(align(8))
#endif


// traits class for facilitating filter coefficients
template <int N>
class coeff_traits
{
public:
	// These coefficients directly relate to the ones in the foton file by
	// gain, a1, a2, b1, b2
	// foton file is "gain a1 a2 b1 b2"
	typedef std::array<double, 4 * N + 1> coeff_array;

	// wide coeffients
	template <int W>
	class coeff_wide_traits
	{
	public:
		typedef alignit(8 * W) std::array<double, W * (4 * N + 1)> coeff_wide_array;
	};

	// compute coefficient a11 of biquad representation
	static constexpr double coeff_a11(double a1, double a2, double b1, double b2)
	{
		return -a1 - 1.0;
	}
	// compute coefficient a12 of biquad representation
	static constexpr double coeff_a12(double a1, double a2, double b1, double b2)
	{
		return -a2 - a1 - 1.0;
	}
	// compute coefficient c1 of biquad representation
	static constexpr double coeff_c1(double a1, double a2, double b1, double b2)
	{
		return b1 - a1;
	}
	// compute coefficient c2 of biquad representation
	static constexpr double coeff_c2(double a1, double a2, double b1, double b2)
	{
		return b2 - a2 + b1 - a1;
	}
	// These coefficients relate to the ones in the foton file by
	// a11 = -a1 - 1
	// a12 = -a2 - a1 - 1
	// c1 = b1 - a1
	// c2 = b2 - a2 + b1 - a1
	// foton file is "gain a1 a2 b1 b2"
	static constexpr coeff_array biquad_shuffle(const coeff_array& coeff)
	{
		coeff_array ret{};
		ret[0] = coeff[0];
		for (int i = 0; i < N; ++i) {
			ret[4 * i + 1] = coeff_a11(coeff[4 * i + 1], coeff[4 * i + 2], coeff[4 * i + 3], coeff[4 * i + 4]);
			ret[4 * i + 2] = coeff_a12(coeff[4 * i + 1], coeff[4 * i + 2], coeff[4 * i + 3], coeff[4 * i + 4]);
			ret[4 * i + 3] = coeff_c1(coeff[4 * i + 1], coeff[4 * i + 2], coeff[4 * i + 3], coeff[4 * i + 4]);
			ret[4 * i + 4] = coeff_c2(coeff[4 * i + 1], coeff[4 * i + 2], coeff[4 * i + 3], coeff[4 * i + 4]);
		}
		return ret;
	}
	// Arrange coefficients W wide
	// W is determined by the number of arguments!
	template <typename... coefarr>
	static constexpr typename coeff_wide_traits<sizeof...(coefarr)>::coeff_wide_array wide(const coefarr&... coeff)
	{
		const size_t W = sizeof...(coefarr);
		typedef std::array<coeff_array, W> arrayargs;
		const arrayargs args{ coeff... };

		typename coeff_wide_traits<W>::coeff_wide_array ret{};
		for (int i = 0; i < 4 * N + 1; ++i)
		{
			for (size_t w = 0; w < W; ++w)
			{
				ret[W * i + w] = args[w][i];
			}
		}
		return ret;
	}
};
