#ifdef __GNUC__
#include <cpuid.h>
#else
#include <intrin.h>
#endif
#include <iostream>
#include <string.h>
#include "instructionset.h"


// Initialize static member data
const InstructionSet::InstructionSet_Internal InstructionSet::CPU_Rep;

// InstructionSet_Internal constructor
InstructionSet::InstructionSet_Internal::InstructionSet_Internal()
    : nIds_{ 0 },
    nExIds_{ 0 },
    isIntel_{ false },
    isAMD_{ false },
    f_1_ECX_{ 0 },
    f_1_EDX_{ 0 },
    f_7_EBX_{ 0 },
    f_7_ECX_{ 0 },
    f_81_ECX_{ 0 },
    f_81_EDX_{ 0 },
    data_{},
    extdata_{}
{
    //int cpuInfo[4] = {-1};
    std::array<int, 4> cpui;

    // Calling __cpuid with 0x0 as the function_id argument
    // gets the number of the highest valid function ID.
#ifdef __GNUC__
    nIds_ = __get_cpuid_max(0, nullptr);
#else
    __cpuid(cpui.data(), 0);
    nIds_ = cpui[0];
#endif
    for (int i = 0; i <= nIds_; ++i)
    {
#ifdef __GNUC__
        __get_cpuid_count (i, 0, (unsigned int*)&cpui[0], (unsigned int*)&cpui[1], (unsigned int*)&cpui[2], (unsigned int*)&cpui[3]);
#else
        __cpuidex(cpui.data(), i, 0);
#endif
        data_.push_back(cpui);
    }

    // Capture vendor string
    char vendor[0x20];
    memset(vendor, 0, sizeof(vendor));
    *reinterpret_cast<int*>(vendor) = data_[0][1];
    *reinterpret_cast<int*>(vendor + 4) = data_[0][3];
    *reinterpret_cast<int*>(vendor + 8) = data_[0][2];

    vendor_ = vendor;
    if (vendor_ == "GenuineIntel")
    {
        isIntel_ = true;
    }
    else if (vendor_ == "AuthenticAMD")
    {
        isAMD_ = true;
    }

    // load bitset with flags for function 0x00000001
    if (nIds_ >= 1)
    {
        f_1_ECX_ = data_[1][2];
        f_1_EDX_ = data_[1][3];
    }

    // load bitset with flags for function 0x00000007
    if (nIds_ >= 7)
    {
        f_7_EBX_ = data_[7][1];
        f_7_ECX_ = data_[7][2];
    }

    // Calling __cpuid with 0x80000000 as the function_id argument
    // gets the number of the highest valid extended ID.
#ifdef __GNUC__
    nExIds_ = __get_cpuid_max(0x80000000, nullptr);
#else
	__cpuid(cpui.data(), 0x80000000);
    nExIds_ = cpui[0];
#endif

    char brand[0x40];
    memset(brand, 0, sizeof(brand));

    for (int i = 0x80000000; i <= nExIds_; ++i)
    {
#ifdef __GNUC__
        __get_cpuid_count (i, 0, (unsigned int*)&cpui[0], (unsigned int*)&cpui[1], (unsigned int*)&cpui[2], (unsigned int*)&cpui[3]);
#else    
        __cpuidex(cpui.data(), i, 0);
#endif        
        extdata_.push_back(cpui);
    }

    // load bitset with flags for function 0x80000001
    if (nExIds_ >= (int)0x80000001)
    {
        f_81_ECX_ = extdata_[1][2];
        f_81_EDX_ = extdata_[1][3];
    }

    // Interpret CPU brand string if reported
    if (nExIds_ >= (int)0x80000004)
    {
        memcpy(brand, extdata_[2].data(), sizeof(cpui));
        memcpy(brand + 16, extdata_[3].data(), sizeof(cpui));
        memcpy(brand + 32, extdata_[4].data(), sizeof(cpui));
        brand_ = brand;
    }
};

// Print out supported instruction set extensions
void InstructionSet::PRINT()
{
    auto& outstream = std::cout;

    auto support_message = [&outstream](std::string isa_feature, bool is_supported) {
        outstream << isa_feature << (is_supported ? " supported" : " not supported") << std::endl;
    };

    std::cout << InstructionSet::Vendor() << std::endl;
    std::cout << InstructionSet::Brand() << std::endl;

    support_message("3DNOW", InstructionSet::_3DNOW());
    support_message("3DNOWEXT", InstructionSet::_3DNOWEXT());
    support_message("ABM", InstructionSet::ABM());
    support_message("ADX", InstructionSet::ADX());
    support_message("AES", InstructionSet::AES());
    support_message("AVX", InstructionSet::AVX());
    support_message("AVX2", InstructionSet::AVX2());
    support_message("AVX512F", InstructionSet::AVX512F());
    support_message("AVX512BW", InstructionSet::AVX512BW());
    support_message("AVX512DQ", InstructionSet::AVX512DQ());
    support_message("AVX512VL", InstructionSet::AVX512VL());
    support_message("AVX512CD", InstructionSet::AVX512CD());
    support_message("AVX512ER", InstructionSet::AVX512ER());
    support_message("AVX512PF", InstructionSet::AVX512PF());
    support_message("BMI1", InstructionSet::BMI1());
    support_message("BMI2", InstructionSet::BMI2());
    support_message("CLFSH", InstructionSet::CLFSH());
    support_message("CMPXCHG16B", InstructionSet::CMPXCHG16B());
    support_message("CX8", InstructionSet::CX8());
    support_message("ERMS", InstructionSet::ERMS());
    support_message("F16C", InstructionSet::F16C());
    support_message("FMA", InstructionSet::FMA());
    support_message("FSGSBASE", InstructionSet::FSGSBASE());
    support_message("FXSR", InstructionSet::FXSR());
    support_message("HLE", InstructionSet::HLE());
    support_message("INVPCID", InstructionSet::INVPCID());
    support_message("LAHF", InstructionSet::LAHF());
    support_message("LZCNT", InstructionSet::LZCNT());
    support_message("MMX", InstructionSet::MMX());
    support_message("MMXEXT", InstructionSet::MMXEXT());
    support_message("MONITOR", InstructionSet::MONITOR());
    support_message("MOVBE", InstructionSet::MOVBE());
    support_message("MSR", InstructionSet::MSR());
    support_message("OSXSAVE", InstructionSet::OSXSAVE());
    support_message("PCLMULQDQ", InstructionSet::PCLMULQDQ());
    support_message("POPCNT", InstructionSet::POPCNT());
    support_message("PREFETCHWT1", InstructionSet::PREFETCHWT1());
    support_message("RDRAND", InstructionSet::RDRAND());
    support_message("RDSEED", InstructionSet::RDSEED());
    support_message("RDTSCP", InstructionSet::RDTSCP());
    support_message("RTM", InstructionSet::RTM());
    support_message("SEP", InstructionSet::SEP());
    support_message("SHA", InstructionSet::SHA());
    support_message("SSE", InstructionSet::SSE());
    support_message("SSE2", InstructionSet::SSE2());
    support_message("SSE3", InstructionSet::SSE3());
    support_message("SSE4.1", InstructionSet::SSE41());
    support_message("SSE4.2", InstructionSet::SSE42());
    support_message("SSE4a", InstructionSet::SSE4a());
    support_message("SSSE3", InstructionSet::SSSE3());
    support_message("SYSCALL", InstructionSet::SYSCALL());
    support_message("TBM", InstructionSet::TBM());
    support_message("XOP", InstructionSet::XOP());
    support_message("XSAVE", InstructionSet::XSAVE());
}

// Print out supported instruction set extensions
void InstructionSet::PRINTAVX()
{
    auto& outstream = std::cout;

    auto support_message = [&outstream](std::string isa_feature, bool is_supported) {
        outstream << isa_feature << (is_supported ? " supported" : " not supported") << std::endl;
    };

    std::cout << InstructionSet::Vendor() << std::endl;
    std::cout << InstructionSet::Brand() << std::endl;

    //support_message("AVX", InstructionSet::AVX());
    support_message("FMA3", InstructionSet::FMA());
    support_message("AVX2", InstructionSet::AVX2());
    support_message("AVX512F", InstructionSet::AVX512F());
    //support_message("AVX512BW", InstructionSet::AVX512BW());
    //support_message("AVX512DQ", InstructionSet::AVX512DQ());
    //support_message("AVX512VL", InstructionSet::AVX512VL());
    //support_message("AVX512CD", InstructionSet::AVX512CD());
}
